/*
Copyright (C) 2000,2001 Stefan Duffner 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <math.h>
#include <qcursor.h>
#include "ICheck.h"
#include "Machine.h"
#include "GState.h"
#include "GTransition.h"
#include "TransitionInfo.h"
#include "IOInfoBin.h"
#include "MainWindow.h"


/// Constructor
ICheck::ICheck(QWidget* m)
{
  main = (MainWindow*)m;
  icheckdlg = new ICheckDlgImpl(m);
}

/// Destructor
ICheck::~ICheck()
{
  delete icheckdlg;
}


/// Performs the whole set of checks and updates the dialog
int ICheck::checkMachine(Machine* m)
{
  if (m==NULL)
    return -1;

  m->updateDefaultTransitions();

  double comp, sreach, esreach, connect;
  QString stmp;

  main->setCursor(Qt::WaitCursor);

  icheckdlg->setUnambiguous(checkUnambigousCond(m));

  icheckdlg->setStart(checkStartState(m));
  icheckdlg->setEnd((checkEndState(m)));
  icheckdlg->setNoDeadLocks(checkDeadLocks(m));

  if (m->getNumInputs()<=31)
  {
    comp = checkCompleteness(m);
    stmp = QString::number(int(comp*100+0.5)) + "%";
  }
  else
    stmp = "?";

  if (m->getType()==Text)
  {
    icheckdlg->setComplete("");
    icheckdlg->enableCompleteness(FALSE);
  }
  else
  {
    icheckdlg->enableCompleteness(TRUE);
    icheckdlg->setComplete(stmp);
  }

  sreach = checkStatesReachable(m);
  stmp = QString::number(int(sreach*100+0.5)) + "%";
  icheckdlg->setStatesReachable(stmp);

  esreach = checkEndStatesReachable(m);
  stmp = QString::number(int(esreach*100+0.5)) + "%";
  icheckdlg->setEndStatesReachable(stmp);

  connect = checkConnections(m);
  stmp = QString::number(int(connect*100+0.5)) + "%";
  icheckdlg->setTransConnected(stmp);

  main->setCursor(Qt::ArrowCursor);

  return 0;
}


/// Checks if there are ambigous conditions
bool ICheck::checkUnambigousCond(Machine* m)
{
  GState* s;
  GTransition *t, *ttmp;
  TransitionInfo *info, *infotmp;
  

  QListIterator<GState> sit(m->getSList());
  for(; sit.current(); ++sit)
  {
    s = sit.current();

    if (!s->isDeleted())
    {
      QListIterator<GTransition> tit(s->tlist);
      for(; tit.current(); ++tit)
      {
	t = tit.current();
	if (!t->isDeleted())
	{
	  info = t->getInfo();

	  QListIterator<GTransition> tit2(s->tlist);
	  for(; tit2.current(); ++tit2)
	  {
	    ttmp = tit2.current();
	    if (!ttmp->isDeleted())
	    {
	      infotmp = ttmp->getInfo();
	      if (infotmp!=info)
	      {
		if (info->intersection(infotmp))
		  return FALSE;
	      }
	    }
	  }
	}
      }
    }
  }
  return TRUE;
}

/// Checks if the machine @a m has a start state
bool ICheck::checkStartState(Machine* m)
{
  if (m->getInitialState()==NULL)
    return FALSE;
  return TRUE;
}

/// Checks if the machine @a m has at least one end state
bool ICheck::checkEndState(Machine* m)
{
  QListIterator<GState> it(m->getSList());
  GState* s;
  for(; it.current(); ++it)
  {
    s = it.current();
    if (!s->isDeleted() && s->isEndState())
      return TRUE;
  }
  return FALSE;
}

/// Checks if there are dead locks in the machine @a m
bool ICheck::checkDeadLocks(Machine* m)
{
  GState *s, *es;
  QList<GState> elist = m->getEndStates();
  elist.setAutoDelete(FALSE);

  QListIterator<GState> it(m->getSList());
  for(; it.current(); ++it)
  {
    s = it.current();

    if (!s->isDeleted())
    {
      bool deadlock=TRUE;
      QListIterator<GState> eit(elist);
      while (eit.current() && deadlock)
      {
	es = eit.current();
	QList<GState> visited, ivisited;
	if (isStateReachable(m, s, es, &visited) || 
	    !isStateReachable(m, m->getInitialState(), s, &ivisited))
	  deadlock=FALSE;
	++eit;
      }

      // check loops of non-end states

      if (!s->isEndState() && m->getNumInputs()<=31)
      {
	bool loop=TRUE;
	int i=0;
	int comb = (int)pow(2, m->getNumInputs());
	IOInfo *io_out;

	while (i<comb && loop)
	{
	  IOInfoBin iobin(IO_MealyIn, i, m->getNumInputs());
	  if (s->next(&iobin, io_out)!=s)
	    loop=FALSE;
	  i++;
	}
	if (loop)
	  deadlock=TRUE;
      }

      if (deadlock)
	return FALSE;
    }
  }
  return TRUE;
}


/// Checks if the machine @a m is complete (i.e. all possible input values are covered)
double ICheck::checkCompleteness(Machine* m)
{
  double rate;
  int match=0, nonmatch=0;
  GState* s;
  IOInfo* io_out;

  if (m->countStates()<2)
    return 1;

  QListIterator<GState> si(m->getSList());
  for(; si.current(); ++si)
  {
    s = si.current();

    if (!s->isDeleted())
    {
      int in;
      in = (int)pow(2, m->getNumInputs());
      for(int i=0; i<in; i++)
      {
	IOInfoBin iobin(IO_MealyIn, i, m->getNumInputs());
	if (s->next(&iobin, io_out))
	  match++;
	else
	  nonmatch++;
      }
    }
  }
  
  if (match+nonmatch>0)
    rate = (double)match/(match+nonmatch);
  else
    rate = 1;
  return rate;
}

/// Checks if all states of the machine @a m are reachable
double ICheck::checkStatesReachable(Machine* m)
{
  int reach=0, nonreach=0;
  double rate;
  GState* istate;
  GState* s;

  istate = m->getInitialState();

  if (m->countStates()==0)
    return 1;

  if (!istate)
    return 0;

  QListIterator<GState> it(m->getSList());
  for(; it.current(); ++it)
  {
    s = it.current();
    QList<GState> visited;
    if (!s->isDeleted())
    {
      if (isStateReachable(m, istate, it.current(), &visited))
	reach++;
      else
	nonreach++;
    }
  }

  if (reach+nonreach>0)
    rate = (double)reach/(reach+nonreach);
  else
    rate = 1;
  return rate;
}


/// Checks if all end states of the machine @a m are reachable
double ICheck::checkEndStatesReachable(Machine* m)
{
  int reach=0, nonreach=0;
  double rate;
  GState* istate;
  GState* s;

  istate = m->getInitialState();

  if (m->countStates()==0)
    return 1;

  if (!istate)
    return 0;

  QListIterator<GState> it(m->getSList());
  for(; it.current(); ++it)
  {
    s = it.current();

    if (!s->isDeleted() && s->isEndState())
    {
      QList<GState> visited;
      if (isStateReachable(m, istate, it.current(), &visited))
	reach++;
      else 
	nonreach++;
    }
  }

  if (reach+nonreach>0)
    rate = (double)reach/(reach+nonreach);
  else
    rate = 1;
  return rate;
}

/// Checks if all all transitions starting from a state have an end state
double ICheck::checkConnections(Machine* m)
{
  int connect=0, nonconnect=0;
  double rate;
  QListIterator<GState> it(m->getSList());
  GState* s;
  GTransition* t;

  for(; it.current(); ++it)
  {
    s = it.current();
    if (!s->isDeleted())
    {
      QListIterator<GTransition> jt(s->tlist);
      for(; jt.current(); ++jt)
      {
	t = jt.current();
	if (t->getEnd()==NULL)
	  nonconnect++;
	else
	  connect++;
      }
    }
  }

  s = m->getPhantomState();
  QListIterator<GTransition> jt(s->tlist);
  for(; jt.current(); ++jt)
  {
    t = jt.current();
//    if (t->getEnd()!=NULL)  
      nonconnect++;
  }

  if (connect+nonconnect>0)
    rate = (double)connect/(nonconnect+connect);
  else
    rate = 1;
  return rate;
}



/**
 * Checks recursively if a state is reachable from another state
 *
 * @param m Machine that contains the states
 * @param from Starting state
 * @param state State that is to be checked for reachability
 * @param visited List of visited states.
 * @returns TRUE if state @a state is reachable from state @a from.
 */
bool ICheck::isStateReachable(Machine* m, GState* from, GState* state, 
    QList<GState>* visited)
{
  GTransition* t;
  GState* s;

  visited->setAutoDelete(FALSE);

  if (!from || !state)
    return FALSE;

  if (from==state)
    return TRUE;

//  QListIterator<GState> it(m->getSList());
  visited->append(state);

  QListIterator<GTransition> it(state->reflist);
  for(; it.current(); ++it)
  {
    t = it.current();
    if (!t->isDeleted())
    {
      s = (GState*)t->getStart();
      if (s && s!=m->getPhantomState() && !s->isDeleted() && !visited->contains(s))
      {
	if (isStateReachable(m, from, s, visited))
	  return TRUE;
      }
    }
  }
  return FALSE;
}


